from typing import List, Optional

from descope._http_base import HTTPBase
from descope.management.common import MgmtV1


class FGA(HTTPBase):

    def __init__(self, http_client, fga_cache_url: Optional[str] = None):
        super().__init__(http_client)
        self._fga_cache_url = fga_cache_url

    def save_schema(self, schema: str):
        """
        Create or update an FGA schema.
        Args:
        schema (str): the schema in the AuthZ 1.0 DSL
            model AuthZ 1.0

            type user

            type org
            relation member: user
            relation parent: org

            type folder
            relation parent: folder
            relation owner: user | org#member
            relation editor: user
            relation viewer: user

            permission can_create: owner | parent.owner
            permission can_edit: editor | can_create
            permission can_view: viewer | can_edit

            type doc
            relation parent: folder
            relation owner: user | org#member
            relation editor: user
            relation viewer: user

            permission can_create: owner | parent.owner
            permission can_edit: editor | can_create
            permission can_view: viewer | can_edit
        Raise:
        AuthException: raised if saving fails
        """
        self._http.post(
            MgmtV1.fga_save_schema,
            body={"dsl": schema},
            base_url=self._fga_cache_url,
        )

    def create_relations(
        self,
        relations: List[dict],
    ):
        """
        Create the given relations based on the existing schema
        Args:
        relations (List[dict]): the relations to create. Each in the following format:
            {
                "resource": "id of the resource that has the relation",
                "resourceType": "the type of the resource (namespace)",
                "relation": "the relation definition for the relation",
                "target": "the target that has the relation - usually users or other resources",
                "targetType": "the type of the target (namespace) - can also be group#member for target sets"
            }
        Raise:
        AuthException: raised if create relations fails
        """
        self._http.post(
            MgmtV1.fga_create_relations,
            body={"tuples": relations},
            base_url=self._fga_cache_url,
        )

    def delete_relations(
        self,
        relations: List[dict],
    ):
        """
        Delete the given relations based on the existing schema
        Args:
        relations (List[dict]): the relations to create. Each in the format as specified above for (create_relations)
        Raise:
        AuthException: raised if delete relations fails
        """
        self._http.post(
            MgmtV1.fga_delete_relations,
            body={"tuples": relations},
            base_url=self._fga_cache_url,
        )

    def check(
        self,
        relations: List[dict],
    ) -> List[dict]:
        """
        Queries the given relations to see if they exist returning true if they do
        Args:
        relations (List[dict]): List of relation queries each in the format of:
            {
                "resource": "id of the resource that has the relation",
                "resourceType": "the type of the resource (namespace)",
                "relation": "the relation definition for the relation",
                "target": "the target that has the relation - usually users or other resources",
                "targetType": "the type of the target (namespace)"
            }

        Return value (List[dict]):
        Return List in the format
             [
                {
                    "allowed": True|False
                    "relation": {
                        "resource": "id of the resource that has the relation",
                        "resourceType": "the type of the resource (namespace)",
                        "relation": "the relation definition for the relation",
                        "target": "the target that has the relation - usually users or other resources",
                        "targetType": "the type of the target (namespace)"
                    }
                }
            ]
        Raise:
        AuthException: raised if query fails
        """
        response = self._http.post(
            MgmtV1.fga_check,
            body={"tuples": relations},
            base_url=self._fga_cache_url,
        )
        return list(
            map(
                lambda tuple: {"relation": tuple["tuple"], "allowed": tuple["allowed"]},
                response.json()["tuples"],
            )
        )

    def load_resources_details(self, resource_identifiers: List[dict]) -> List[dict]:
        """
        Load details for the given resource identifiers.
        Args:
            resource_identifiers (List[dict]): list of dicts each containing 'resourceId' and 'resourceType'.
        Returns:
            List[dict]: list of resources details as returned by the server.
        """
        response = self._http.post(
            MgmtV1.fga_resources_load,
            body={"resourceIdentifiers": resource_identifiers},
        )
        return response.json().get("resourcesDetails", [])

    def save_resources_details(self, resources_details: List[dict]) -> None:
        """
        Save details for the given resources.
        Args:
            resources_details (List[dict]): list of dicts each containing 'resourceId' and 'resourceType' plus optionally containing metadata fields such as 'displayName'.
        """
        self._http.post(
            MgmtV1.fga_resources_save,
            body={"resourcesDetails": resources_details},
        )



from __future__ import annotations

from typing import Any, List, Optional

from descope._http_base import HTTPBase
from descope.management.common import (
    MgmtV1,
    SAMLIDPAttributeMappingInfo,
    SAMLIDPGroupsMappingInfo,
    saml_idp_attribute_mapping_info_to_dict,
    saml_idp_groups_mapping_info_to_dict,
)


class SSOApplication(HTTPBase):
    def create_oidc_application(
        self,
        name: str,
        login_page_url: str,
        id: Optional[str] = None,
        description: Optional[str] = None,
        logo: Optional[str] = None,
        enabled: Optional[bool] = True,
        force_authentication: Optional[bool] = False,
    ) -> dict:
        """
        Create a new OIDC sso application with the given name. SSO application IDs are provisioned automatically, but can be provided
        explicitly if needed. Both the name and ID must be unique per project.

        Args:
        name (str): The sso application's name.
        login_page_url (str): The URL where login page is hosted.
        id (str): Optional sso application ID.
        description (str): Optional sso application description.
        logo (str): Optional sso application logo.
        enabled (bool): Optional (default True) does the sso application will be enabled or disabled.
        force_authentication (bool): Optional determine if the IdP should force the user to re-authenticate.

        Return value (dict):
        Return dict in the format
             {"id": <id>}

        Raise:
        AuthException: raised if create operation fails
        """
        uri = MgmtV1.sso_application_oidc_create_path
        response = self._http.post(
            uri,
            body=SSOApplication._compose_create_update_oidc_body(
                name,
                login_page_url,
                id,
                description,
                logo,
                enabled,
                force_authentication,
            ),
        )
        return response.json()

    def create_saml_application(
        self,
        name: str,
        login_page_url: str,
        id: Optional[str] = None,
        description: Optional[str] = None,
        logo: Optional[str] = None,
        enabled: Optional[bool] = True,
        use_metadata_info: Optional[bool] = False,
        metadata_url: Optional[str] = None,
        entity_id: Optional[str] = None,
        acs_url: Optional[str] = None,
        certificate: Optional[str] = None,
        attribute_mapping: Optional[List[SAMLIDPAttributeMappingInfo]] = None,
        groups_mapping: Optional[List[SAMLIDPGroupsMappingInfo]] = None,
        acs_allowed_callbacks: Optional[List[str]] = None,
        subject_name_id_type: Optional[str] = None,
        subject_name_id_format: Optional[str] = None,
        default_relay_state: Optional[str] = None,
        force_authentication: Optional[bool] = False,
        logout_redirect_url: Optional[str] = None,
    ) -> dict:
        """
        Create a new SAML sso application with the given name. SSO application IDs are provisioned automatically, but can be provided
        explicitly if needed. Both the name and ID must be unique per project.

        Args:
        name (str): The sso application's name.
        login_page_url (str): The URL where login page is hosted.
        id (str): Optional sso application ID.
        description (str): Optional sso application description.
        logo (str): Optional sso application logo.
        enabled (bool): Optional set the sso application as enabled or disabled.
        use_metadata_info (bool): Optional determine if SP info should be automatically fetched from metadata_url or by specified it by the entity_id, acs_url, certificate parameters.
        metadata_url (str): Optional SP metadata url which include all the SP SAML info.
        entity_id (str): Optional SP entity id.
        acs_url (str): Optional SP ACS (saml callback) url.
        certificate (str): Optional SP certificate, relevant only when SAML request must be signed.
        attribute_mapping (List[SAMLIDPAttributeMappingInfo]): Optional list of Descope (IdP) attributes to SP mapping.
        groups_mapping (List[SAMLIDPGroupsMappingInfo]): Optional list of Descope (IdP) roles that will be mapped to SP groups.
        acs_allowed_callbacks (List[str]): Optional list of urls wildcards strings represents the allowed ACS urls that will be accepted while arriving on the SAML request as SP callback urls.
        subject_name_id_type (str): Optional define the SAML Assertion subject name type, leave empty for using Descope user-id or set to "email"/"phone".
        subject_name_id_format (str): Optional define the SAML Assertion subject name format, leave empty for using "urn:oasis:names:tc:SAML:1.1:nameid-format:unspecified".
        default_relay_state (str): Optional define the default relay state.
        force_authentication (bool): Optional determine if the IdP should force the user to re-authenticate.
        logout_redirect_url (str): Optional Target URL to which the user will be redirected upon logout completion.

        Return value (dict):
        Return dict in the format
             {"id": <id>}

        Raise:
        AuthException: raised if create operation fails
        """

        if use_metadata_info:
            if not metadata_url:
                raise Exception("metadata_url argument must be set")
        else:
            if not entity_id or not acs_url or not certificate:
                raise Exception("entity_id, acs_url, certificate arguments must be set")

        attribute_mapping = [] if attribute_mapping is None else attribute_mapping

        groups_mapping = [] if groups_mapping is None else groups_mapping

        acs_allowed_callbacks = (
            [] if acs_allowed_callbacks is None else acs_allowed_callbacks
        )

        uri = MgmtV1.sso_application_saml_create_path
        response = self._http.post(
            uri,
            body=SSOApplication._compose_create_update_saml_body(
                name,
                login_page_url,
                id,
                description,
                enabled,
                logo,
                use_metadata_info,
                metadata_url,
                entity_id,
                acs_url,
                certificate,
                attribute_mapping,
                groups_mapping,
                acs_allowed_callbacks,
                subject_name_id_type,
                subject_name_id_format,
                default_relay_state,
                force_authentication,
                logout_redirect_url,
            ),
        )
        return response.json()

    def update_oidc_application(
        self,
        id: str,
        name: str,
        login_page_url: str,
        description: Optional[str] = None,
        logo: Optional[str] = None,
        enabled: Optional[bool] = True,
        force_authentication: Optional[bool] = False,
    ):
        """
        Update an existing OIDC sso application with the given parameters. IMPORTANT: All parameters are used as overrides
        to the existing sso application. Empty fields will override populated fields. Use carefully.

        Args:
        id (str): The ID of the sso application to update.
        name (str): Updated sso application name
        login_page_url (str): The URL where login page is hosted.
        description (str): Optional sso application description.
        logo (str): Optional sso application logo.
        enabled (bool): Optional (default True) does the sso application will be enabled or disabled.
        force_authentication (bool): Optional determine if the IdP should force the user to re-authenticate.

        Raise:
        AuthException: raised if update operation fails
        """

        uri = MgmtV1.sso_application_oidc_update_path
        self._http.post(
            uri,
            body=SSOApplication._compose_create_update_oidc_body(
                name,
                login_page_url,
                id,
                description,
                logo,
                enabled,
                force_authentication,
            ),
        )

    def update_saml_application(
        self,
        id: str,
        name: str,
        login_page_url: str,
        description: Optional[str] = None,
        logo: Optional[str] = None,
        enabled: Optional[bool] = True,
        use_metadata_info: Optional[bool] = False,
        metadata_url: Optional[str] = None,
        entity_id: Optional[str] = None,
        acs_url: Optional[str] = None,
        certificate: Optional[str] = None,
        attribute_mapping: Optional[List[SAMLIDPAttributeMappingInfo]] = None,
        groups_mapping: Optional[List[SAMLIDPGroupsMappingInfo]] = None,
        acs_allowed_callbacks: Optional[List[str]] = None,
        subject_name_id_type: Optional[str] = None,
        subject_name_id_format: Optional[str] = None,
        default_relay_state: Optional[str] = None,
        force_authentication: Optional[bool] = False,
        logout_redirect_url: Optional[str] = None,
    ):
        """
        Update an existing SAML sso application with the given parameters. IMPORTANT: All parameters are used as overrides
        to the existing sso application. Empty fields will override populated fields. Use carefully.

        Args:
        id (str): The ID of the sso application to update.
        name (str): Updated sso application name
        login_page_url (str): The URL where login page is hosted.
        description (str): Optional sso application description.
        logo (str): Optional sso application logo.
        enabled (bool): Optional (default True) does the sso application will be enabled or disabled.
        use_metadata_info (bool): Optional determine if SP info should be automatically fetched from metadata_url or by specified it by the entity_id, acs_url, certificate parameters.
        metadata_url (str): Optional SP metadata url which include all the SP SAML info.
        entity_id (str): Optional SP entity id.
        acs_url (str): Optional SP ACS (saml callback) url.
        certificate (str): Optional SP certificate, relevant only when SAML request must be signed.
        attribute_mapping (List[SAMLIDPAttributeMappingInfo]): Optional list of Descope (IdP) attributes to SP mapping.
        groups_mapping (List[SAMLIDPGroupsMappingInfo]): Optional list of Descope (IdP) roles that will be mapped to SP groups.
        acs_allowed_callbacks (List[str]): Optional list of urls wildcards strings represents the allowed ACS urls that will be accepted while arriving on the SAML request as SP callback urls.
        subject_name_id_type (str): Optional define the SAML Assertion subject name type, leave empty for using Descope user-id or set to "email"/"phone".
        subject_name_id_format (str): Optional define the SAML Assertion subject name format, leave empty for using "urn:oasis:names:tc:SAML:1.1:nameid-format:unspecified".
        default_relay_state (str): Optional define the default relay state.
        force_authentication (bool): Optional determine if the IdP should force the user to re-authenticate.
        logout_redirect_url (str): Optional Target URL to which the user will be redirected upon logout completion.

        Raise:
        AuthException: raised if update operation fails
        """

        if use_metadata_info:
            if not metadata_url:
                raise Exception("metadata_url argument must be set")
        else:
            if not entity_id or not acs_url or not certificate:
                raise Exception("entity_id, acs_url, certificate arguments must be set")

        attribute_mapping = [] if attribute_mapping is None else attribute_mapping

        groups_mapping = [] if groups_mapping is None else groups_mapping

        acs_allowed_callbacks = (
            [] if acs_allowed_callbacks is None else acs_allowed_callbacks
        )

        uri = MgmtV1.sso_application_saml_update_path
        self._http.post(
            uri,
            body=SSOApplication._compose_create_update_saml_body(
                name,
                login_page_url,
                id,
                description,
                enabled,
                logo,
                use_metadata_info,
                metadata_url,
                entity_id,
                acs_url,
                certificate,
                attribute_mapping,
                groups_mapping,
                acs_allowed_callbacks,
                subject_name_id_type,
                subject_name_id_format,
                default_relay_state,
                force_authentication,
                logout_redirect_url,
            ),
        )

    def delete(
        self,
        id: str,
    ):
        """
        Delete an existing sso application. IMPORTANT: This action is irreversible. Use carefully.

        Args:
        id (str): The ID of the sso application that's to be deleted.

        Raise:
        AuthException: raised if deletion operation fails
        """
        uri = MgmtV1.sso_application_delete_path
        # Using adapter's do_post which already includes management key in Authorization header
        self._http.post(uri, body={"id": id})

    def load(
        self,
        id: str,
    ) -> dict:
        """
        Load sso application by id.

        Args:
        id (str): The ID of the sso application to load.

        Return value (dict):
        Return dict in the format
             {"id":"<id>","name":"<name>","description":"<description>","enabled":true,"logo":"","appType":"saml","samlSettings":{"loginPageUrl":"","idpCert":"<cert>","useMetadataInfo":true,"metadataUrl":"","entityId":"","acsUrl":"","certificate":"","attributeMapping":[{"name":"email","type":"","value":"attrVal1"}],"groupsMapping":[{"name":"grp1","type":"","filterType":"roles","value":"","roles":[{"id":"myRoleId","name":"myRole"}]}],"idpMetadataUrl":"","idpEntityId":"","idpSsoUrl":"","acsAllowedCallbacks":[],"subjectNameIdType":"","subjectNameIdFormat":"", "defaultRelayState":"", "forceAuthentication": false, "idpLogoutUrl": "", "logoutRedirectUrl": ""},"oidcSettings":{"loginPageUrl":"","issuer":"","discoveryUrl":"", "forceAuthentication":false}}
        Containing the loaded sso application information.

        Raise:
        AuthException: raised if load operation fails
        """
        response = self._http.get(MgmtV1.sso_application_load_path, params={"id": id})
        return response.json()

    def load_all(
        self,
    ) -> dict:
        """
        Load all sso applications.

        Return value (dict):
        Return dict in the format
             {
                                "apps": [
                                        {"id":"app1","name":"<name>","description":"<description>","enabled":true,"logo":"","appType":"saml","samlSettings":{"loginPageUrl":"","idpCert":"<cert>","useMetadataInfo":true,"metadataUrl":"","entityId":"","acsUrl":"","certificate":"","attributeMapping":[{"name":"email","type":"","value":"attrVal1"}],"groupsMapping":[{"name":"grp1","type":"","filterType":"roles","value":"","roles":[{"id":"myRoleId","name":"myRole"}]}],"idpMetadataUrl":"","idpEntityId":"","idpSsoUrl":"","acsAllowedCallbacks":[],"subjectNameIdType":"","subjectNameIdFormat":"", "defaultRelayState":"", "forceAuthentication": false, "idpLogoutUrl": "", "logoutRedirectUrl": ""},"oidcSettings":{"loginPageUrl":"","issuer":"","discoveryUrl":"", "forceAuthentication":false}},
                                {"id":"app2","name":"<name>","description":"<description>","enabled":true,"logo":"","appType":"saml","samlSettings":{"loginPageUrl":"","idpCert":"<cert>","useMetadataInfo":true,"metadataUrl":"","entityId":"","acsUrl":"","certificate":"","attributeMapping":[{"name":"email","type":"","value":"attrVal1"}],"groupsMapping":[{"name":"grp1","type":"","filterType":"roles","value":"","roles":[{"id":"myRoleId","name":"myRole"}]}],"idpMetadataUrl":"","idpEntityId":"","idpSsoUrl":"","acsAllowedCallbacks":[],"subjectNameIdType":"","subjectNameIdFormat":"", "defaultRelayState":"", "forceAuthentication": false, "idpLogoutUrl": "", "logoutRedirectUrl": ""},"oidcSettings":{"loginPageUrl":"","issuer":"","discoveryUrl":"", "forceAuthentication":false}}
                                ]
            }
        Containing the loaded sso applications information.

        Raise:
        AuthException: raised if load operation fails
        """
        response = self._http.get(MgmtV1.sso_application_load_all_path)
        return response.json()

    @staticmethod
    def _compose_create_update_oidc_body(
        name: str,
        login_page_url: str,
        id: Optional[str] = None,
        description: Optional[str] = None,
        logo: Optional[str] = None,
        enabled: Optional[bool] = True,
        force_authentication: Optional[bool] = False,
    ) -> dict:
        body: dict[str, Any] = {
            "name": name,
            "id": id,
            "description": description,
            "logo": logo,
            "enabled": enabled,
            "loginPageUrl": login_page_url,
            "forceAuthentication": force_authentication,
        }
        return body

    @staticmethod
    def _compose_create_update_saml_body(
        name: str,
        login_page_url: str,
        id: Optional[str] = None,
        description: Optional[str] = None,
        enabled: Optional[bool] = True,
        logo: Optional[str] = None,
        use_metadata_info: Optional[bool] = False,
        metadata_url: Optional[str] = None,
        entity_id: Optional[str] = None,
        acs_url: Optional[str] = None,
        certificate: Optional[str] = None,
        attribute_mapping: Optional[List[SAMLIDPAttributeMappingInfo]] = None,
        groups_mapping: Optional[List[SAMLIDPGroupsMappingInfo]] = None,
        acs_allowed_callbacks: Optional[List[str]] = None,
        subject_name_id_type: Optional[str] = None,
        subject_name_id_format: Optional[str] = None,
        default_relay_state: Optional[str] = None,
        force_authentication: Optional[bool] = False,
        logout_redirect_url: Optional[str] = None,
    ) -> dict:
        body: dict[str, Any] = {
            "id": id,
            "name": name,
            "description": description,
            "enabled": enabled,
            "logo": logo,
            "loginPageUrl": login_page_url,
            "useMetadataInfo": use_metadata_info,
            "metadataUrl": metadata_url,
            "entityId": entity_id,
            "acsUrl": acs_url,
            "certificate": certificate,
            "attributeMapping": saml_idp_attribute_mapping_info_to_dict(
                attribute_mapping
            ),
            "groupsMapping": saml_idp_groups_mapping_info_to_dict(groups_mapping),
            "acsAllowedCallbacks": acs_allowed_callbacks,
            "subjectNameIdType": subject_name_id_type,
            "subjectNameIdFormat": subject_name_id_format,
            "defaultRelayState": default_relay_state,
            "forceAuthentication": force_authentication,
            "logoutRedirectUrl": logout_redirect_url,
        }

        return body



from __future__ import annotations

from typing import Iterable

from descope._auth_base import AuthBase
from descope.auth import Auth
from descope.common import (
    REFRESH_SESSION_COOKIE_NAME,
    DeliveryMethod,
    EndpointsV1,
    LoginOptions,
    SignUpOptions,
    signup_options_to_dict,
    validate_refresh_token_provided,
)
from descope.exceptions import ERROR_TYPE_INVALID_ARGUMENT, AuthException


class OTP(AuthBase):
    def sign_in(
        self,
        method: DeliveryMethod,
        login_id: str,
        login_options: LoginOptions | None = None,
        refresh_token: str | None = None,
    ) -> str:
        """
        Sign in (log in) an existing user with the unique login_id you provide. (See 'sign_up' function for an explanation of the
            login_id field.) Provide the DeliveryMethod required for this user. If the login_id value cannot be used for the
            DeliverMethod selected (for example, 'login_id = 4567qq445km' and 'DeliveryMethod = email')

        Args:
        method (DeliveryMethod): The method to use for delivering the OTP verification code to the user, for example
            Email, SMS, Voice call, or WhatsApp
        login_id (str): The login ID of the user being validated for example phone or email
        login_options (LoginOptions): Optional advanced controls over login parameters
        refresh_token: Optional refresh token is needed for specific login options

        Raise:
        AuthException: raised if sign-in operation fails
        """
        if not login_id:
            raise AuthException(
                400, ERROR_TYPE_INVALID_ARGUMENT, "Identifier cannot be empty"
            )

        validate_refresh_token_provided(login_options, refresh_token)

        uri = OTP._compose_signin_url(method)
        body = OTP._compose_signin_body(login_id, login_options)
        response = self._http.post(uri, body=body, pswd=refresh_token)
        return Auth.extract_masked_address(response.json(), method)

    def sign_up(
        self,
        method: DeliveryMethod,
        login_id: str,
        user: dict | None = None,
        signup_options: SignUpOptions | None = None,
    ) -> str:
        """
        Sign up (create) a new user using their email or phone number. Choose a delivery method for OTP
            verification, for example Email, SMS, Voice call, or WhatsApp.
            (optional) Include additional user metadata that you wish to preserve.

        Args:
        method (DeliveryMethod): The method to use for delivering the OTP verification code, for example phone or email
        login_id (str): The login ID of the user being validated
        user (dict) optional: Preserve additional user metadata in the form of
             {"name": "Joe Person", "phone": "2125551212", "email": "joe@somecompany.com"}

        Raise:
        AuthException: raised if sign-up operation fails
        """

        if not user:
            user = {}

        if not self._auth.adjust_and_verify_delivery_method(method, login_id, user):
            raise AuthException(
                400,
                ERROR_TYPE_INVALID_ARGUMENT,
                f"Login ID {login_id} is not valid by delivery method {method}",
            )

        uri = OTP._compose_signup_url(method)
        body = OTP._compose_signup_body(method, login_id, user, signup_options)
        response = self._http.post(uri, body=body)
        return Auth.extract_masked_address(response.json(), method)

    def sign_up_or_in(
        self,
        method: DeliveryMethod,
        login_id: str,
        signup_options: SignUpOptions | None = None,
    ) -> str:
        """
        Sign_up_or_in lets you handle both sign up and sign in with a single call. Sign-up_or_in will first determine if
            login_id is a new or existing end user. If login_id is new, a new end user user will be created and then
            authenticated using the OTP DeliveryMethod specified. If login_id exists, the end user will be authenticated
            using the OTP DeliveryMethod specified.

        Args:
        method (DeliveryMethod): The method to use for delivering the OTP verification code, for example Email, SMS, Voice call, or WhatsApp
        login_id (str): The Login ID of the user being validated

        Raise:
        AuthException: raised if either the sign_up or sign_in operation fails
        """
        if not login_id:
            raise AuthException(
                400, ERROR_TYPE_INVALID_ARGUMENT, "Identifier cannot be empty"
            )

        uri = OTP._compose_sign_up_or_in_url(method)
        login_options: LoginOptions | None = None
        if signup_options is not None:
            login_options = LoginOptions(
                custom_claims=signup_options.customClaims,
                template_options=signup_options.templateOptions,
                template_id=signup_options.templateId,
            )
        body = OTP._compose_signin_body(
            login_id,
            login_options,
        )
        response = self._http.post(uri, body=body)
        return Auth.extract_masked_address(response.json(), method)

    def verify_code(
        self,
        method: DeliveryMethod,
        login_id: str,
        code: str,
        audience: str | None | Iterable[str] = None,
    ) -> dict:
        """
        Verify the validity of an OTP code entered by an end user during sign_in or sign_up.
        (This function is not needed if you are using the sign_up_or_in function.

        Args:
        method (DeliveryMethod): The method to use for delivering the OTP verification code, for example Email, SMS, Voice call, or WhatsApp
        login_id (str): The Login ID of the user being validated
        code (str): The authorization code enter by the end user during signup/signin

        Return value (dict):
        Return dict in the format
             {"jwts": [], "user": "", "firstSeen": "", "error": ""}
        Includes all the jwts tokens (session token, refresh token), token claims, and user information

        Raise:
        AuthException: raised if the OTP code is not valid or if token verification failed
        """
        if not login_id:
            raise AuthException(
                400, ERROR_TYPE_INVALID_ARGUMENT, "Identifier cannot be empty"
            )

        uri = OTP._compose_verify_code_url(method)
        body = OTP._compose_verify_code_body(login_id, code)
        response = self._http.post(uri, body=body)

        resp = response.json()
        jwt_response = self._auth.generate_jwt_response(
            resp, response.cookies.get(REFRESH_SESSION_COOKIE_NAME, None), audience
        )
        return jwt_response

    def update_user_email(
        self,
        login_id: str,
        email: str,
        refresh_token: str,
        add_to_login_ids: bool = False,
        on_merge_use_existing: bool = False,
        template_options: dict | None = None,
        template_id: str | None = None,
        provider_id: str | None = None,
    ) -> str:
        """
        Update the email address of an end user, after verifying the authenticity of the end user using OTP.

        Args:
        login_id (str): The login ID of the user whose information is being updated
        email (str): The new email address. If an email address already exists for this end user, it will be overwritten
        refresh_token (str): The session's refresh token (used for verification)

        Raise:
        AuthException: raised if OTP verification fails or if token verification fails
        """

        if not login_id:
            raise AuthException(
                400, ERROR_TYPE_INVALID_ARGUMENT, "Identifier cannot be empty"
            )

        Auth.validate_email(email)

        uri = EndpointsV1.update_user_email_otp_path
        body = OTP._compose_update_user_email_body(
            login_id,
            email,
            add_to_login_ids,
            on_merge_use_existing,
            template_options,
            template_id,
            provider_id,
        )
        response = self._http.post(uri, body=body, pswd=refresh_token)
        return Auth.extract_masked_address(response.json(), DeliveryMethod.EMAIL)

    def update_user_phone(
        self,
        method: DeliveryMethod,
        login_id: str,
        phone: str,
        refresh_token: str,
        add_to_login_ids: bool = False,
        on_merge_use_existing: bool = False,
        template_options: dict | None = None,
        template_id: str | None = None,
        provider_id: str | None = None,
    ) -> str:
        """
        Update the phone number of an existing end user, after verifying the authenticity of the end user using OTP.

        Args:
        method (DeliveryMethod): The method to use for delivering the OTP verification code, for example Email, SMS, Voice call, or WhatsApp
        login_id (str): The login ID of the user whose information is being updated
        phone (str): The new phone number. If a phone number already exists for this end user, it will be overwritten
        refresh_token (str): The session's refresh token (used for OTP verification)
        add_to_login_ids (bool): Defaults to false, determine whether to add this email to the login ids of hte user or not
        on_merge_use_existing (bool): Defaults to false, In case add_to_login_ids and there is such a user already
            determine whether keep the existing user, or this new one

        Raise:
        AuthException: raised if OTP verification fails or if token verification fails
        """

        if not login_id:
            raise AuthException(
                400, ERROR_TYPE_INVALID_ARGUMENT, "Identifier cannot be empty"
            )

        Auth.validate_phone(method, phone)

        uri = OTP._compose_update_phone_url(method)
        body = OTP._compose_update_user_phone_body(
            login_id,
            phone,
            add_to_login_ids,
            on_merge_use_existing,
            template_options,
            template_id,
            provider_id,
        )
        response = self._http.post(uri, body=body, pswd=refresh_token)
        return Auth.extract_masked_address(response.json(), method)

    @staticmethod
    def _compose_signup_url(method: DeliveryMethod) -> str:
        return Auth.compose_url(EndpointsV1.sign_up_auth_otp_path, method)

    @staticmethod
    def _compose_signin_url(method: DeliveryMethod) -> str:
        return Auth.compose_url(EndpointsV1.sign_in_auth_otp_path, method)

    @staticmethod
    def _compose_sign_up_or_in_url(method: DeliveryMethod) -> str:
        return Auth.compose_url(EndpointsV1.sign_up_or_in_auth_otp_path, method)

    @staticmethod
    def _compose_verify_code_url(method: DeliveryMethod) -> str:
        return Auth.compose_url(EndpointsV1.verify_code_auth_path, method)

    @staticmethod
    def _compose_update_phone_url(method: DeliveryMethod) -> str:
        return Auth.compose_url(EndpointsV1.update_user_phone_otp_path, method)

    @staticmethod
    def _compose_signup_body(
        method: DeliveryMethod,
        login_id: str,
        user: dict,
        signup_options: SignUpOptions | None = None,
    ) -> dict:
        body: dict[str, str | bool | dict] = {"loginId": login_id}

        if signup_options is not None:
            body["loginOptions"] = signup_options_to_dict(signup_options)

        if user is not None:
            body["user"] = user
            method_str, val = Auth.get_login_id_by_method(method, user)
            body[method_str] = val
        return body

    @staticmethod
    def _compose_signin_body(
        login_id: str, login_options: LoginOptions | None = None
    ) -> dict:
        return {
            "loginId": login_id,
            "loginOptions": login_options.__dict__ if login_options else {},
        }

    @staticmethod
    def _compose_verify_code_body(login_id: str, code: str) -> dict:
        return {"loginId": login_id, "code": code}

    @staticmethod
    def _compose_update_user_email_body(
        login_id: str,
        email: str,
        add_to_login_ids: bool,
        on_merge_use_existing: bool,
        template_options: dict | None = None,
        template_id: str | None = None,
        provider_id: str | None = None,
    ) -> dict:
        body: dict[str, str | bool | dict] = {
            "loginId": login_id,
            "email": email,
            "addToLoginIDs": add_to_login_ids,
            "onMergeUseExisting": on_merge_use_existing,
        }
        if template_options is not None:
            body["templateOptions"] = template_options
        if template_id is not None:
            body["templateId"] = template_id
        if provider_id is not None:
            body["providerId"] = provider_id

        return body

    @staticmethod
    def _compose_update_user_phone_body(
        login_id: str,
        phone: str,
        add_to_login_ids: bool,
        on_merge_use_existing: bool,
        template_options: dict | None = None,
        template_id: str | None = None,
        provider_id: str | None = None,
    ) -> dict:
        body: dict[str, str | bool | dict] = {
            "loginId": login_id,
            "phone": phone,
            "addToLoginIDs": add_to_login_ids,
            "onMergeUseExisting": on_merge_use_existing,
        }
        if template_options is not None:
            body["templateOptions"] = template_options
        if template_id is not None:
            body["templateId"] = template_id
        if provider_id is not None:
            body["providerId"] = provider_id

        return body



from __future__ import annotations

from typing import Any, List, Optional

from descope._http_base import HTTPBase
from descope.exceptions import ERROR_TYPE_INVALID_ARGUMENT, AuthException  # noqa: F401
from descope.http_client import HTTPClient
from descope.management.common import (
    AccessType,
    MgmtV1,
    PromptType,
    URLParam,
    url_params_to_dict,
)


class _OutboundApplicationTokenFetcher:
    """Internal helper class for shared token fetching logic."""

    @staticmethod
    def fetch_token_by_scopes(
        *,
        http: HTTPClient,
        token: Optional[str] = None,
        app_id: str,
        user_id: str,
        scopes: List[str],
        options: Optional[dict] = None,
        tenant_id: Optional[str] = None,
    ) -> dict:
        """Internal implementation for fetching token by scopes."""
        uri = MgmtV1.outbound_application_fetch_token_by_scopes_path
        response = http.post(
            uri,
            body={
                "appId": app_id,
                "userId": user_id,
                "scopes": scopes,
                "options": options,
                "tenantId": tenant_id,
            },
            pswd=token,
        )
        return response.json()

    @staticmethod
    def fetch_token(
        *,
        http: HTTPClient,
        token: Optional[str] = None,
        app_id: str,
        user_id: str,
        tenant_id: Optional[str] = None,
        options: Optional[dict] = None,
    ) -> dict:
        """Internal implementation for fetching token."""
        uri = MgmtV1.outbound_application_fetch_token_path
        response = http.post(
            uri,
            body={
                "appId": app_id,
                "userId": user_id,
                "tenantId": tenant_id,
                "options": options,
            },
            pswd=token,
        )
        return response.json()

    @staticmethod
    def fetch_tenant_token_by_scopes(
        *,
        http: HTTPClient,
        token: Optional[str] = None,
        app_id: str,
        tenant_id: str,
        scopes: List[str],
        options: Optional[dict] = None,
    ) -> dict:
        """Internal implementation for fetching tenant token by scopes."""
        uri = MgmtV1.outbound_application_fetch_tenant_token_by_scopes_path
        response = http.post(
            uri,
            body={
                "appId": app_id,
                "tenantId": tenant_id,
                "scopes": scopes,
                "options": options,
            },
            pswd=token,
        )
        return response.json()

    @staticmethod
    def fetch_tenant_token(
        *,
        http: HTTPClient,
        token: Optional[str] = None,
        app_id: str,
        tenant_id: str,
        options: Optional[dict] = None,
    ) -> dict:
        """Internal implementation for fetching tenant token."""
        uri = MgmtV1.outbound_application_fetch_tenant_token_path
        response = http.post(
            uri,
            body={
                "appId": app_id,
                "tenantId": tenant_id,
                "options": options,
            },
            pswd=token,
        )
        return response.json()


class OutboundApplication(HTTPBase):
    def create_application(
        self,
        name: str,
        description: Optional[str] = None,
        logo: Optional[str] = None,
        id: Optional[str] = None,
        client_secret: Optional[str] = None,
        client_id: Optional[str] = None,
        discovery_url: Optional[str] = None,
        authorization_url: Optional[str] = None,
        authorization_url_params: Optional[List[URLParam]] = None,
        token_url: Optional[str] = None,
        token_url_params: Optional[List[URLParam]] = None,
        revocation_url: Optional[str] = None,
        default_scopes: Optional[List[str]] = None,
        default_redirect_url: Optional[str] = None,
        callback_domain: Optional[str] = None,
        pkce: Optional[bool] = None,
        access_type: Optional[AccessType] = None,
        prompt: Optional[List[PromptType]] = None,
    ) -> dict:
        """
        Create a new outbound application with the given name. Outbound application IDs are provisioned automatically, but can be provided
        explicitly if needed. Both the name and ID must be unique per project.

        Args:
        name (str): The outbound application's name.
        description (str): Optional outbound application description.
        logo (str): Optional outbound application logo.
        id (str): Optional outbound application ID.
        client_secret (str): Optional client secret for the application.
        client_id (str): Optional client ID for the application.
        discovery_url (str): Optional OAuth discovery URL.
        authorization_url (str): Optional OAuth authorization URL.
        authorization_url_params (List[URLParam]): Optional authorization URL parameters.
        token_url (str): Optional OAuth token URL.
        token_url_params (List[URLParam]): Optional token URL parameters.
        revocation_url (str): Optional OAuth token revocation URL.
        default_scopes (List[str]): Optional default OAuth scopes.
        default_redirect_url (str): Optional default redirect URL.
        callback_domain (str): Optional callback domain.
        pkce (bool): Optional PKCE (Proof Key for Code Exchange) support.
        access_type (AccessType): Optional OAuth access type.
        prompt (List[PromptType]): Optional OAuth prompt parameters.

        Return value (dict):
        Return dict in the format
             {"app": {"id": <id>, "name": <name>, "description": <description>, "logo": <logo>}}

        Raise:
        AuthException: raised if create operation fails
        """
        uri = MgmtV1.outbound_application_create_path
        response = self._http.post(
            uri,
            body=OutboundApplication._compose_create_update_body(
                name,
                description,
                logo,
                id,
                client_secret,
                client_id,
                discovery_url,
                authorization_url,
                authorization_url_params,
                token_url,
                token_url_params,
                revocation_url,
                default_scopes,
                default_redirect_url,
                callback_domain,
                pkce,
                access_type,
                prompt,
            ),
        )
        return response.json()

    def update_application(
        self,
        id: str,
        name: str,
        description: Optional[str] = None,
        logo: Optional[str] = None,
        client_secret: Optional[str] = None,
        client_id: Optional[str] = None,
        discovery_url: Optional[str] = None,
        authorization_url: Optional[str] = None,
        authorization_url_params: Optional[List[URLParam]] = None,
        token_url: Optional[str] = None,
        token_url_params: Optional[List[URLParam]] = None,
        revocation_url: Optional[str] = None,
        default_scopes: Optional[List[str]] = None,
        default_redirect_url: Optional[str] = None,
        callback_domain: Optional[str] = None,
        pkce: Optional[bool] = None,
        access_type: Optional[AccessType] = None,
        prompt: Optional[List[PromptType]] = None,
    ) -> dict:
        """
        Update an existing outbound application with the given parameters. IMPORTANT: All parameters are used as overrides
        to the existing outbound application. Empty fields will override populated fields. Use carefully.

        Args:
        id (str): The ID of the outbound application to update.
        name (str): Updated outbound application name.
        description (str): Optional outbound application description.
        logo (str): Optional outbound application logo.
        client_secret (str): Optional client secret for the application.
        client_id (str): Optional client ID for the application.
        discovery_url (str): Optional OAuth discovery URL.
        authorization_url (str): Optional OAuth authorization URL.
        authorization_url_params (List[URLParam]): Optional authorization URL parameters.
        token_url (str): Optional OAuth token URL.
        token_url_params (List[URLParam]): Optional token URL parameters.
        revocation_url (str): Optional OAuth token revocation URL.
        default_scopes (List[str]): Optional default OAuth scopes.
        default_redirect_url (str): Optional default redirect URL.
        callback_domain (str): Optional callback domain.
        pkce (bool): Optional PKCE (Proof Key for Code Exchange) support.
        access_type (AccessType): Optional OAuth access type.
        prompt (List[PromptType]): Optional OAuth prompt parameters.

        Return value (dict):
        Return dict in the format
             {"app": {"id": <id>, "name": <name>, "description": <description>, "logo": <logo>}}

        Raise:
        AuthException: raised if update operation fails
        """
        uri = MgmtV1.outbound_application_update_path
        response = self._http.post(
            uri,
            body={
                "app": OutboundApplication._compose_create_update_body(
                    name,
                    description,
                    logo,
                    id,
                    client_secret,
                    client_id,
                    discovery_url,
                    authorization_url,
                    authorization_url_params,
                    token_url,
                    token_url_params,
                    revocation_url,
                    default_scopes,
                    default_redirect_url,
                    callback_domain,
                    pkce,
                    access_type,
                    prompt,
                )
            },
        )
        return response.json()

    def delete_application(
        self,
        id: str,
    ):
        """
        Delete an existing outbound application. IMPORTANT: This action is irreversible. Use carefully.

        Args:
        id (str): The ID of the outbound application that's to be deleted.

        Raise:
        AuthException: raised if deletion operation fails
        """
        uri = MgmtV1.outbound_application_delete_path
        self._http.post(uri, body={"id": id})

    def load_application(
        self,
        id: str,
    ) -> dict:
        """
        Load outbound application by id.

        Args:
        id (str): The ID of the outbound application to load.

        Return value (dict):
        Return dict in the format
             {"app": {"id": <id>, "name": <name>, "description": <description>, "logo": <logo>}}
        Containing the loaded outbound application information.

        Raise:
        AuthException: raised if load operation fails
        """
        response = self._http.get(f"{MgmtV1.outbound_application_load_path}/{id}")
        return response.json()

    def load_all_applications(self) -> dict:
        """
        Load all outbound applications.

        Return value (dict):
        Return dict in the format
             {"apps": [{"id": <id>, "name": <name>, "description": <description>, "logo": <logo>}, ...]}
        Containing the loaded outbound applications information.

        Raise:
        AuthException: raised if load operation fails
        """
        response = self._http.get(MgmtV1.outbound_application_load_all_path)
        return response.json()

    def fetch_token_by_scopes(
        self,
        app_id: str,
        user_id: str,
        scopes: List[str],
        options: Optional[dict] = None,
        tenant_id: Optional[str] = None,
    ) -> dict:
        """
        Fetch an outbound application token for a user with specific scopes.

        Args:
        app_id (str): The ID of the outbound application.
        user_id (str): The ID of the user.
        scopes (List[str]): List of scopes to include in the token.
        options (dict): Optional token options.
        tenant_id (str): Optional tenant ID.

        Return value (dict):
        Return dict in the format
             {"token": {"token": <access_token>, "refreshToken": <refresh_token>, "expiresIn": <expires_in>, "tokenType": <token_type>, "scopes": <scopes>}}

        Raise:
        AuthException: raised if fetch operation fails
        """
        return _OutboundApplicationTokenFetcher.fetch_token_by_scopes(
            http=self._http,
            app_id=app_id,
            user_id=user_id,
            scopes=scopes,
            options=options,
            tenant_id=tenant_id,
        )

    def fetch_token(
        self,
        app_id: str,
        user_id: str,
        tenant_id: Optional[str] = None,
        options: Optional[dict] = None,
    ) -> dict:
        """
        Fetch an outbound application token for a user.

        Args:
        app_id (str): The ID of the outbound application.
        user_id (str): The ID of the user.
        tenant_id (str): Optional tenant ID.
        options (dict): Optional token options.

        Return value (dict):
        Return dict in the format
             {"token": {"token": <access_token>, "refreshToken": <refresh_token>, "expiresIn": <expires_in>, "tokenType": <token_type>, "scopes": <scopes>}}

        Raise:
        AuthException: raised if fetch operation fails
        """
        return _OutboundApplicationTokenFetcher.fetch_token(
            http=self._http,
            app_id=app_id,
            user_id=user_id,
            tenant_id=tenant_id,
            options=options,
        )

    def fetch_tenant_token_by_scopes(
        self,
        app_id: str,
        tenant_id: str,
        scopes: List[str],
        options: Optional[dict] = None,
    ) -> dict:
        """
        Fetch an outbound application token for a tenant with specific scopes.

        Args:
        app_id (str): The ID of the outbound application.
        tenant_id (str): The ID of the tenant.
        scopes (List[str]): List of scopes to include in the token.
        options (dict): Optional token options.

        Return value (dict):
        Return dict in the format
             {"token": {"token": <access_token>, "refreshToken": <refresh_token>, "expiresIn": <expires_in>, "tokenType": <token_type>, "scopes": <scopes>}}

        Raise:
        AuthException: raised if fetch operation fails
        """
        return _OutboundApplicationTokenFetcher.fetch_tenant_token_by_scopes(
            http=self._http,
            app_id=app_id,
            tenant_id=tenant_id,
            scopes=scopes,
            options=options,
        )

    def fetch_tenant_token(
        self,
        app_id: str,
        tenant_id: str,
        options: Optional[dict] = None,
    ) -> dict:
        """
        Fetch an outbound application token for a tenant.

        Args:
        app_id (str): The ID of the outbound application.
        tenant_id (str): The ID of the tenant.
        options (dict): Optional token options.

        Return value (dict):
        Return dict in the format
             {"token": {"token": <access_token>, "refreshToken": <refresh_token>, "expiresIn": <expires_in>, "tokenType": <token_type>, "scopes": <scopes>}}

        Raise:
        AuthException: raised if fetch operation fails
        """
        return _OutboundApplicationTokenFetcher.fetch_tenant_token(
            http=self._http,
            app_id=app_id,
            tenant_id=tenant_id,
            options=options,
        )

    @staticmethod
    def _compose_create_update_body(
        name: str,
        description: Optional[str] = None,
        logo: Optional[str] = None,
        id: Optional[str] = None,
        client_secret: Optional[str] = None,
        client_id: Optional[str] = None,
        discovery_url: Optional[str] = None,
        authorization_url: Optional[str] = None,
        authorization_url_params: Optional[List[URLParam]] = None,
        token_url: Optional[str] = None,
        token_url_params: Optional[List[URLParam]] = None,
        revocation_url: Optional[str] = None,
        default_scopes: Optional[List[str]] = None,
        default_redirect_url: Optional[str] = None,
        callback_domain: Optional[str] = None,
        pkce: Optional[bool] = None,
        access_type: Optional[AccessType] = None,
        prompt: Optional[List[PromptType]] = None,
    ) -> dict:
        body: dict[str, Any] = {
            "name": name,
            "id": id,
            "description": description,
            "logo": logo,
        }
        if client_secret:
            body["clientSecret"] = client_secret
        if client_id:
            body["clientId"] = client_id
        if discovery_url:
            body["discoveryUrl"] = discovery_url
        if authorization_url:
            body["authorizationUrl"] = authorization_url
        if authorization_url_params is not None:
            body["authorizationUrlParams"] = url_params_to_dict(
                authorization_url_params
            )
        if token_url:
            body["tokenUrl"] = token_url
        if token_url_params is not None:
            body["tokenUrlParams"] = url_params_to_dict(token_url_params)
        if revocation_url:
            body["revocationUrl"] = revocation_url
        if default_scopes is not None:
            body["defaultScopes"] = default_scopes
        if default_redirect_url:
            body["defaultRedirectUrl"] = default_redirect_url
        if callback_domain:
            body["callbackDomain"] = callback_domain
        if pkce is not None:
            body["pkce"] = pkce
        if access_type:
            body["accessType"] = access_type.value
        if prompt is not None:
            body["prompt"] = [p.value for p in prompt]
        return body


class OutboundApplicationByToken(HTTPBase):
    def __init__(self, http_client: HTTPClient):
        # This class expects the token to be passed for each call
        no_key_client = HTTPClient(
            project_id=http_client.project_id,
            base_url=http_client.base_url,
            timeout_seconds=http_client.timeout_seconds,
            secure=http_client.secure,
            management_key=None,  # Override the management key for this client
        )
        super().__init__(no_key_client)

    # Methods for fetching outbound application tokens using an inbound application token
    # that includes the "outbound.token.fetch" scope (no management key required)

    def _check_inbound_app_token(self, token: str):
        """Check if inbound app token is available for the given property."""
        if not token:
            raise AuthException(
                400,
                ERROR_TYPE_INVALID_ARGUMENT,
                "Inbound app token is required for perform this functionality",
            )

    def fetch_token_by_scopes(
        self,
        token: str,
        app_id: str,
        user_id: str,
        scopes: List[str],
        options: Optional[dict] = None,
        tenant_id: Optional[str] = None,
    ) -> dict:
        """
        Fetch an outbound application token for a user with specific scopes.

        Args:
        token (str): The Inbound Application token to use for authentication.
        app_id (str): The ID of the outbound application.
        user_id (str): The ID of the user.
        scopes (List[str]): List of scopes to include in the token.
        options (dict): Optional token options.
        tenant_id (str): Optional tenant ID.

        Return value (dict):
        Return dict in the format
             {"token": {"token": <access_token>, "refreshToken": <refresh_token>, "expiresIn": <expires_in>, "tokenType": <token_type>, "scopes": <scopes>}}

        Raise:
        AuthException: raised if fetch operation fails
        """
        self._check_inbound_app_token(token)
        return _OutboundApplicationTokenFetcher.fetch_token_by_scopes(
            http=self._http,
            token=token,
            app_id=app_id,
            user_id=user_id,
            scopes=scopes,
            options=options,
            tenant_id=tenant_id,
        )

    def fetch_token(
        self,
        token: str,
        app_id: str,
        user_id: str,
        tenant_id: Optional[str] = None,
        options: Optional[dict] = None,
    ) -> dict:
        """
        Fetch an outbound application token for a user.

        Args:
        token (str): The Inbound Application token to use for authentication.
        app_id (str): The ID of the outbound application.
        user_id (str): The ID of the user.
        tenant_id (str): Optional tenant ID.
        options (dict): Optional token options.

        Return value (dict):
        Return dict in the format
             {"token": {"token": <access_token>, "refreshToken": <refresh_token>, "expiresIn": <expires_in>, "tokenType": <token_type>, "scopes": <scopes>}}

        Raise:
        AuthException: raised if fetch operation fails
        """
        self._check_inbound_app_token(token)
        return _OutboundApplicationTokenFetcher.fetch_token(
            http=self._http,
            token=token,
            app_id=app_id,
            user_id=user_id,
            tenant_id=tenant_id,
            options=options,
        )

    def fetch_tenant_token_by_scopes(
        self,
        token: str,
        app_id: str,
        tenant_id: str,
        scopes: List[str],
        options: Optional[dict] = None,
    ) -> dict:
        """
        Fetch an outbound application token for a tenant with specific scopes.

        Args:
        token (str): The Inbound Application token to use for authentication.
        app_id (str): The ID of the outbound application.
        tenant_id (str): The ID of the tenant.
        scopes (List[str]): List of scopes to include in the token.
        options (dict): Optional token options.

        Return value (dict):
        Return dict in the format
             {"token": {"token": <access_token>, "refreshToken": <refresh_token>, "expiresIn": <expires_in>, "tokenType": <token_type>, "scopes": <scopes>}}

        Raise:
        AuthException: raised if fetch operation fails
        """
        self._check_inbound_app_token(token)
        return _OutboundApplicationTokenFetcher.fetch_tenant_token_by_scopes(
            http=self._http,
            token=token,
            app_id=app_id,
            tenant_id=tenant_id,
            scopes=scopes,
            options=options,
        )

    def fetch_tenant_token(
        self, token: str, app_id: str, tenant_id: str, options: Optional[dict] = None
    ) -> dict:
        """
        Fetch an outbound application token for a tenant.

        Args:
        token (str): The Inbound Application token to use for authentication.
        app_id (str): The ID of the outbound application.
        tenant_id (str): The ID of the tenant.
        options (dict): Optional token options.

        Return value (dict):
        Return dict in the format
             {"token": {"token": <access_token>, "refreshToken": <refresh_token>, "expiresIn": <expires_in>, "tokenType": <token_type>, "scopes": <scopes>}}

        Raise:
        AuthException: raised if fetch operation fails
        """
        self._check_inbound_app_token(token)
        return _OutboundApplicationTokenFetcher.fetch_tenant_token(
            http=self._http,
            token=token,
            app_id=app_id,
            tenant_id=tenant_id,
            options=options,
        )



from enum import Enum
from typing import List, Optional


class SessionExpirationUnit(Enum):
    MINUTES = "minutes"
    HOURS = "hours"
    DAYS = "days"
    WEEKS = "weeks"


class TenantAuthType(Enum):
    NONE = "none"
    SAML = "saml"
    OIDC = "oidc"


class AccessType(Enum):
    OFFLINE = "offline"
    ONLINE = "online"


class PromptType(Enum):
    NONE = "none"
    LOGIN = "login"
    CONSENT = "consent"
    SELECT_ACCOUNT = "select_account"


class URLParam:
    def __init__(self, name: str, value: str):
        self.name = name
        self.value = value

    def to_dict(self) -> dict:
        return {"name": self.name, "value": self.value}


def url_params_to_dict(url_params: Optional[List[URLParam]] = None) -> list:
    if url_params is None:
        return []
    return [param.to_dict() for param in url_params]


class MgmtV1:
    # tenant
    tenant_create_path = "/v1/mgmt/tenant/create"
    tenant_update_path = "/v1/mgmt/tenant/update"
    tenant_delete_path = "/v1/mgmt/tenant/delete"
    tenant_load_path = "/v1/mgmt/tenant"
    tenant_settings_path = "/v1/mgmt/tenant/settings"
    tenant_load_all_path = "/v1/mgmt/tenant/all"
    tenant_search_all_path = "/v1/mgmt/tenant/search"

    # sso application
    sso_application_oidc_create_path = "/v1/mgmt/sso/idp/app/oidc/create"
    sso_application_saml_create_path = "/v1/mgmt/sso/idp/app/saml/create"
    sso_application_oidc_update_path = "/v1/mgmt/sso/idp/app/oidc/update"
    sso_application_saml_update_path = "/v1/mgmt/sso/idp/app/saml/update"
    sso_application_delete_path = "/v1/mgmt/sso/idp/app/delete"
    sso_application_load_path = "/v1/mgmt/sso/idp/app/load"
    sso_application_load_all_path = "/v1/mgmt/sso/idp/apps/load"

    # outbound application
    outbound_application_create_path = "/v1/mgmt/outbound/app/create"
    outbound_application_update_path = "/v1/mgmt/outbound/app/update"
    outbound_application_delete_path = "/v1/mgmt/outbound/app/delete"
    outbound_application_load_path = "/v1/mgmt/outbound/app"
    outbound_application_load_all_path = "/v1/mgmt/outbound/apps"
    outbound_application_fetch_token_by_scopes_path = "/v1/mgmt/outbound/app/user/token"
    outbound_application_fetch_token_path = "/v1/mgmt/outbound/app/user/token/latest"
    outbound_application_fetch_tenant_token_by_scopes_path = (
        "/v1/mgmt/outbound/app/tenant/token"
    )
    outbound_application_fetch_tenant_token_path = (
        "/v1/mgmt/outbound/app/tenant/token/latest"
    )

    # user
    user_create_path = "/v1/mgmt/user/create"
    test_user_create_path = "/v1/mgmt/user/create/test"
    user_create_batch_path = "/v1/mgmt/user/create/batch"
    user_update_path = "/v1/mgmt/user/update"
    user_patch_path = "/v1/mgmt/user/patch"
    user_patch_batch_path = "/v1/mgmt/user/patch/batch"
    user_delete_path = "/v1/mgmt/user/delete"
    user_logout_path = "/v1/mgmt/user/logout"
    user_delete_all_test_users_path = "/v1/mgmt/user/test/delete/all"
    user_load_path = "/v1/mgmt/user"
    users_load_path = "/v1/mgmt/users/load"
    users_search_path = "/v2/mgmt/user/search"
    test_users_search_path = "/v2/mgmt/user/search/test"
    user_get_provider_token = "/v1/mgmt/user/provider/token"
    user_update_status_path = "/v1/mgmt/user/update/status"
    user_update_login_id_path = "/v1/mgmt/user/update/loginid"
    user_update_email_path = "/v1/mgmt/user/update/email"
    user_update_phone_path = "/v1/mgmt/user/update/phone"
    user_update_name_path = "/v1/mgmt/user/update/name"
    user_update_picture_path = "/v1/mgmt/user/update/picture"
    user_update_custom_attribute_path = "/v1/mgmt/user/update/customAttribute"
    user_set_role_path = "/v1/mgmt/user/update/role/set"
    user_add_role_path = "/v2/mgmt/user/update/role/add"
    user_remove_role_path = "/v1/mgmt/user/update/role/remove"
    user_add_sso_apps = "/v1/mgmt/user/update/ssoapp/add"
    user_set_sso_apps = "/v1/mgmt/user/update/ssoapp/set"
    user_remove_sso_apps = "/v1/mgmt/user/update/ssoapp/remove"
    user_set_password_path = "/v1/mgmt/user/password/set"  # Deprecated
    user_set_temporary_password_path = "/v1/mgmt/user/password/set/temporary"
    user_set_active_password_path = "/v1/mgmt/user/password/set/active"
    user_expire_password_path = "/v1/mgmt/user/password/expire"
    user_remove_all_passkeys_path = "/v1/mgmt/user/passkeys/delete"
    user_remove_totp_seed_path = "/v1/mgmt/user/totp/delete"
    user_add_tenant_path = "/v1/mgmt/user/update/tenant/add"
    user_remove_tenant_path = "/v1/mgmt/user/update/tenant/remove"
    user_generate_otp_for_test_path = "/v1/mgmt/tests/generate/otp"
    user_generate_magic_link_for_test_path = "/v1/mgmt/tests/generate/magiclink"
    user_generate_enchanted_link_for_test_path = "/v1/mgmt/tests/generate/enchantedlink"
    user_generate_embedded_link_path = "/v1/mgmt/user/signin/embeddedlink"
    user_generate_sign_up_embedded_link_path = "/v1/mgmt/user/signup/embeddedlink"
    user_history_path = "/v1/mgmt/user/history"

    # access key
    access_key_create_path = "/v1/mgmt/accesskey/create"
    access_key_load_path = "/v1/mgmt/accesskey"
    access_keys_search_path = "/v1/mgmt/accesskey/search"
    access_key_update_path = "/v1/mgmt/accesskey/update"
    access_key_deactivate_path = "/v1/mgmt/accesskey/deactivate"
    access_key_activate_path = "/v1/mgmt/accesskey/activate"
    access_key_delete_path = "/v1/mgmt/accesskey/delete"

    # sso
    sso_settings_path = "/v1/mgmt/sso/settings"
    sso_metadata_path = "/v1/mgmt/sso/metadata"
    sso_mapping_path = "/v1/mgmt/sso/mapping"
    sso_load_settings_path = "/v2/mgmt/sso/settings"  # v2 only
    sso_configure_oidc_settings = "/v1/mgmt/sso/oidc"
    sso_configure_saml_settings = "/v1/mgmt/sso/saml"
    sso_configure_saml_by_metadata_settings = "/v1/mgmt/sso/saml/metadata"

    # jwt
    update_jwt_path = "/v1/mgmt/jwt/update"
    impersonate_path = "/v1/mgmt/impersonate"
    stop_impersonation_path = "/v1/mgmt/stop/impersonation"
    mgmt_sign_in_path = "/v1/mgmt/auth/signin"
    mgmt_sign_up_path = "/v1/mgmt/auth/signup"
    mgmt_sign_up_or_in_path = "/v1/mgmt/auth/signup-in"
    anonymous_path = "/v1/mgmt/auth/anonymous"

    # permission
    permission_create_path = "/v1/mgmt/permission/create"
    permission_update_path = "/v1/mgmt/permission/update"
    permission_delete_path = "/v1/mgmt/permission/delete"
    permission_load_all_path = "/v1/mgmt/permission/all"

    # role
    role_create_path = "/v1/mgmt/role/create"
    role_update_path = "/v1/mgmt/role/update"
    role_delete_path = "/v1/mgmt/role/delete"
    role_load_all_path = "/v1/mgmt/role/all"
    role_search_path = "/v1/mgmt/role/search"

    # flow
    flow_list_path = "/v1/mgmt/flow/list"
    flow_delete_path = "/v1/mgmt/flow/delete"
    flow_import_path = "/v1/mgmt/flow/import"
    flow_export_path = "/v1/mgmt/flow/export"

    # theme
    theme_import_path = "/v1/mgmt/theme/import"
    theme_export_path = "/v1/mgmt/theme/export"

    # group
    group_load_all_path = "/v1/mgmt/group/all"
    group_load_all_for_member_path = "/v1/mgmt/group/member/all"
    group_load_all_group_members_path = "/v1/mgmt/group/members"

    # Audit
    audit_search = "/v1/mgmt/audit/search"
    audit_create_event = "/v1/mgmt/audit/event"

    # Authz ReBAC
    authz_schema_save = "/v1/mgmt/authz/schema/save"
    authz_schema_delete = "/v1/mgmt/authz/schema/delete"
    authz_schema_load = "/v1/mgmt/authz/schema/load"
    authz_ns_save = "/v1/mgmt/authz/ns/save"
    authz_ns_delete = "/v1/mgmt/authz/ns/delete"
    authz_rd_save = "/v1/mgmt/authz/rd/save"
    authz_rd_delete = "/v1/mgmt/authz/rd/delete"
    authz_re_create = "/v1/mgmt/authz/re/create"
    authz_re_delete = "/v1/mgmt/authz/re/delete"
    authz_re_delete_resources = "/v1/mgmt/authz/re/deleteresources"
    authz_re_has_relations = "/v1/mgmt/authz/re/has"
    authz_re_who = "/v1/mgmt/authz/re/who"
    authz_re_resource = "/v1/mgmt/authz/re/resource"
    authz_re_targets = "/v1/mgmt/authz/re/targets"
    authz_re_target_all = "/v1/mgmt/authz/re/targetall"
    authz_re_target_with_relation = "/v1/mgmt/authz/re/targetwithrelation"
    authz_get_modified = "/v1/mgmt/authz/getmodified"

    # FGA (new style Authz)
    fga_save_schema = "/v1/mgmt/fga/schema"
    fga_create_relations = "/v1/mgmt/fga/relations"
    fga_delete_relations = "/v1/mgmt/fga/relations/delete"
    fga_check = "/v1/mgmt/fga/check"
    fga_resources_load = "/v1/mgmt/fga/resources/load"
    fga_resources_save = "/v1/mgmt/fga/resources/save"

    # Project
    project_update_name = "/v1/mgmt/project/update/name"
    project_update_tags = "/v1/mgmt/project/update/tags"
    project_clone = "/v1/mgmt/project/clone"
    project_export = "/v1/mgmt/project/export"
    project_import = "/v1/mgmt/project/import"
    project_list_projects = "/v1/mgmt/projects/list"


class MgmtSignUpOptions:
    def __init__(
        self,
        custom_claims: Optional[dict] = None,
        refresh_duration: Optional[int] = None,
    ):
        self.custom_claims = custom_claims
        self.refresh_duration = refresh_duration


class MgmtLoginOptions:
    def __init__(
        self,
        stepup: bool = False,
        mfa: bool = False,
        revoke_other_sessions: Optional[bool] = None,
        custom_claims: Optional[dict] = None,
        jwt: Optional[str] = None,
        refresh_duration: Optional[int] = None,
    ):
        self.stepup = stepup
        self.custom_claims = custom_claims
        self.mfa = mfa
        self.revoke_other_sessions = revoke_other_sessions
        self.jwt = jwt
        self.refresh_duration = refresh_duration


def is_jwt_required(lgo: MgmtLoginOptions) -> bool:
    return lgo is not None and (lgo.stepup or lgo.mfa)


class MgmtUserRequest:
    def __init__(
        self,
        name: Optional[str] = None,
        given_name: Optional[str] = None,
        middle_name: Optional[str] = None,
        family_name: Optional[str] = None,
        phone: Optional[str] = None,
        email: Optional[str] = None,
        email_verified: Optional[bool] = None,
        phone_verified: Optional[bool] = None,
        sso_app_id: Optional[str] = None,
    ):
        self.name = name
        self.given_name = given_name
        self.middle_name = middle_name
        self.family_name = family_name
        self.phone = phone
        self.email = email
        self.email_verified = email_verified
        self.phone_verified = phone_verified
        self.sso_app_id = sso_app_id

    def to_dict(self) -> dict:
        return {
            "name": self.name,
            "givenName": self.given_name,
            "middleName": self.middle_name,
            "familyName": self.family_name,
            "phone": self.phone,
            "email": self.email,
            "emailVerified": self.email_verified,
            "phoneVerified": self.phone_verified,
            "ssoAppId": self.sso_app_id,
        }


class AssociatedTenant:
    """
    Represents a tenant association for a User or Access Key. The tenant_id is required to denote
    which tenant the user or access key belongs to. The role_names array is an optional list of
    roles for the user or access key in this specific tenant.
    """

    def __init__(self, tenant_id: str, role_names: Optional[List[str]] = None):
        self.tenant_id = tenant_id
        self.role_names = [] if role_names is None else role_names


def associated_tenants_to_dict(associated_tenants: List[AssociatedTenant]) -> list:
    associated_tenant_list = []
    if associated_tenants:
        for associated_tenant in associated_tenants:
            associated_tenant_list.append(
                {
                    "tenantId": associated_tenant.tenant_id,
                    "roleNames": associated_tenant.role_names,
                }
            )
    return associated_tenant_list


class SAMLIDPAttributeMappingInfo:
    """
    Represents a SAML IDP attribute mapping object. use this class for mapping Descope attribute
    to the relevant SAML Assertion attributes matching your expected SP attributes names.
    """

    def __init__(self, name: str, type: str, value: str):
        self.name = name
        self.type = type
        self.value = value


def saml_idp_attribute_mapping_info_to_dict(
    attributes_mapping: Optional[List[SAMLIDPAttributeMappingInfo]] = None,
) -> list:
    attributes_mapping_list = []
    if attributes_mapping:
        for attribute_mapping in attributes_mapping:
            attributes_mapping_list.append(
                {
                    "name": attribute_mapping.name,
                    "type": attribute_mapping.type,
                    "value": attribute_mapping.value,
                }
            )
    return attributes_mapping_list


class SAMLIDPRoleGroupMappingInfo:
    """
    Represents a SAML IDP Role Group mapping object.
    """

    def __init__(self, id: str, name: str):
        self.id = id
        self.name = name


def saml_idp_role_group_mapping_info_to_dict(
    role_groups_mapping: Optional[List[SAMLIDPRoleGroupMappingInfo]] = None,
) -> list:
    role_groups_mapping_list = []
    if role_groups_mapping:
        for group_mapping in role_groups_mapping:
            role_groups_mapping_list.append(
                {
                    "id": group_mapping.id,
                    "name": group_mapping.name,
                }
            )
    return role_groups_mapping_list


class SAMLIDPGroupsMappingInfo:
    """
    Represents a SAML IDP Descope roles to SP groups mapping object. use this class for mapping Descope roles
    to your SP groups.
    """

    def __init__(
        self,
        name: str,
        type: str,
        filter_type: str,
        value: str,
        roles: List[SAMLIDPRoleGroupMappingInfo],
    ):
        self.name = name
        self.type = type
        self.filter_type = filter_type
        self.value = value
        self.roles = roles


def saml_idp_groups_mapping_info_to_dict(
    groups_mapping: Optional[List[SAMLIDPGroupsMappingInfo]] = None,
) -> list:
    groups_mapping_list = []
    if groups_mapping:
        for group_mapping in groups_mapping:
            groups_mapping_list.append(
                {
                    "name": group_mapping.name,
                    "type": group_mapping.type,
                    "filterType": group_mapping.filter_type,
                    "value": group_mapping.value,
                    "roles": saml_idp_role_group_mapping_info_to_dict(
                        group_mapping.roles
                    ),
                }
            )
    return groups_mapping_list


class Sort:
    """
    Represents a sort object.
    """

    def __init__(self, field: str, desc: Optional[bool] = False):
        self.field = field
        self.desc = desc


def sort_to_dict(sort: List[Sort]) -> list:
    sort_list = []
    if sort:
        for s in sort:
            sort_list.append(
                {
                    "field": s.field,
                    "desc": s.desc,
                }
            )
    return sort_list
